/*
 * Simple PostgreSQL client library for Lua.
 * Copyright (c) 2007 Michael Broughton (mbobowik@telusplanet.net)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */


#include <stdio.h>

#include "libpq-fe.h"
#include "lauxlib.h"
#include "lua.h"


struct lpg_stmt {
	int	closed;
	int	conn;		/* Reference to PGconn userdata. */
	int	params;		/* Number of parameters to send. */
};


LUALIB_API int luaopen_postgres(lua_State *L);


/*
 * Create a new PGconn userdata.
 */
static PGconn **lpg_newconn(lua_State *L, PGconn *pg_conn)
{
	PGconn **conn = lua_newuserdata(L, sizeof(*conn));
	*conn = pg_conn;

	lua_pushstring(L, "lpg_conn_metatable");
	lua_rawget(L, LUA_ENVIRONINDEX);
	lua_setmetatable(L, -2);

	return conn;
}


/*
 * Check if the object at idx is a valid PGconn userdata.
 * Optionally check if the connection has been free'd. This does not check
 * the status of the actual PGconn.
 * Note: idx must be an absolute index.
 */
static PGconn **lpg_checkconn(lua_State *L, int idx, int check_closed)
{
	PGconn **conn;
	lua_pushstring(L, "lpg_conn_metatable");
	lua_rawget(L, LUA_ENVIRONINDEX);
	lua_getmetatable(L, idx);
	if (lua_rawequal(L, -1, -2) && (conn = lua_touserdata(L, idx))) {
		lua_pop(L, 2);
		if (check_closed && !*conn)
			luaL_error(L, "PostgreSQL connection closed.");
		return conn;
	}
	luaL_argerror(L, idx, "PostgreSQL connection expected");
	return NULL;
}


/*
 * Create a new statement userdata. This references a PGconn userdata.
 * The environment table of this object is used to store the statement
 * string, as well as its parameters and parameter types.
 * Note: conn_ud and statement must be absolute indicies.
 */
static struct lpg_stmt *lpg_newstmt(lua_State *L, int conn_ud, int statement)
{
	struct lpg_stmt *stmt = lua_newuserdata(L, sizeof(*stmt));
	stmt->closed = 0;
	stmt->conn = LUA_NOREF;
	stmt->params = 0;

	lua_pushvalue(L, conn_ud);
	stmt->conn = luaL_ref(L, LUA_ENVIRONINDEX);

	lua_pushstring(L, "lpg_stmt_metatable");
	lua_rawget(L, LUA_ENVIRONINDEX);
	lua_setmetatable(L, -2);

	lua_createtable(L, 0, 1);
	lua_pushstring(L, "statement");
	lua_pushvalue(L, statement);
	lua_rawset(L, -3);
	lua_setfenv(L, -2);

	return stmt;
}


/*
 * Check if the object at idx is a valid statement userdata.
 * Optionally check if the statement has been closed.
 * Note: idx must be an absolute index.
 */
static struct lpg_stmt *lpg_checkstmt(lua_State *L, int idx, int check_closed)
{
	struct lpg_stmt *stmt;
	lua_pushstring(L, "lpg_stmt_metatable");
	lua_rawget(L, LUA_ENVIRONINDEX);
	lua_getmetatable(L, idx);
	if (lua_rawequal(L, -1, -2) && (stmt = lua_touserdata(L, idx))) {
		lua_pop(L, 2);
		if (check_closed && stmt->closed)
			luaL_error(L, "PostgreSQL statement closed.");
		return stmt;
	}
	luaL_argerror(L, idx, "PostgreSQL statement expected");
	return NULL;
}


/****************************************************************************/


/*
 * Execute a statement with or without a parameters on the given PGconn.
 * Returns a PGresult, which must be free'd by the caller.
 * Note: param_table must be an absolute index.
 * Param table:
 *	Indicies 1 .. n hold the parameter strings.
 *	Indicies -1 .. -n hold the parameter type numbers (Oid's).
 */
static PGresult *lpg_execute(lua_State *L, PGconn *pg_conn,
				const char *statement, int params,
				int param_table)
{
	if (params > 0) {
		const char *param_vals[params];
		Oid param_types[params];
		int i;
		for (i = 0; i < params; i++) {
			lua_rawgeti(L, param_table, i + 1);
			lua_rawgeti(L, param_table, -(i + 1));
			if (lua_type(L, -2) == LUA_TSTRING)
				param_vals[i] = lua_tostring(L, -2);
			else
				param_vals[i] = NULL;
			if (lua_type(L, -1) == LUA_TNUMBER)
				param_types[i] = (Oid)lua_tonumber(L, -1);
			else
				param_types[i] = 0;
			lua_pop(L, 2);
		}
		return PQexecParams(pg_conn, statement, params, param_types,
					param_vals, NULL, NULL, 0);
	} else
		return PQexec(pg_conn, statement);
}


/*
 * Push the status of a PGresult onto the stack and free the PGresult.
 */
static int lpg_res_clear(lua_State *L, PGresult *res)
{
	if (res && PQresultStatus(res) == PGRES_COMMAND_OK) {
		lua_pushstring(L, PQcmdTuples(res));
		lua_pushnumber(L, lua_tonumber(L, -1));
		lua_remove(L, -2);
		PQclear(res);
		return 1;
	} else if (res && PQresultStatus(res) == PGRES_TUPLES_OK) {
		lua_pushinteger(L, PQntuples(res));
		PQclear(res);
		return 1;
	} else {
		lua_pushnil(L);
		lua_pushstring(L, PQresultErrorMessage(res));
		PQclear(res);
		return 2;
	}
}


/*
 * Pop a data type string from the top of the stack and replace it with its
 * corresponding Oid.
 */
static void lpg_getoid(lua_State *L, PGconn *pg_conn)
{
	int t = lua_gettop(L);
	/* Check the Oid cache for the type string. */
	lua_pushstring(L, "lpg_oid_cache");
	lua_rawget(L, LUA_ENVIRONINDEX);
	lua_pushvalue(L, t);
	lua_rawget(L, -2);
	if (lua_type(L, -1) != LUA_TNUMBER) {
		/* Cache miss. Look up the Oid in the DB. */
		PGresult *res;
		char *stmt = "SELECT oid FROM pg_type WHERE typname = $1::text";
		const char *typ;
		lua_pop(L, 1);
		typ = lua_tostring(L, t);
		res = PQexecParams(pg_conn, stmt, 1, NULL, &typ, NULL, NULL, 0);
		if (res && PQresultStatus(res) == PGRES_TUPLES_OK &&
			PQntuples(res) == 1 && !PQgetisnull(res, 0, 0)) {
			/* An Oid was found. */
			lua_pushstring(L, PQgetvalue(res, 0, 0));
			lua_pushnumber(L, lua_tonumber(L, -1));
			lua_remove(L, -2);
			lua_pushvalue(L, t);
			/* Cache the type string and Oid. */
			lua_pushvalue(L, -2);
			lua_pushvalue(L, -2);
			lua_rawset(L, t + 1);
			lua_pushvalue(L, -2);
			lua_rawset(L, t + 1);
		} else
			/* An Oid was not found push nil in its place. */
			lua_pushnil(L);
		PQclear(res);
	}
	lua_remove(L, -2);
	lua_remove(L, -2);
}


/*
 * Push 'value' onto the stack. If there is a type parser availible for the
 * given 'type', run the 'value' string through it and push the result onto
 * the stack instead.
 */
static void lpg_pushtype(lua_State *L, PGconn *pg_conn, Oid type,
				const char *value)
{
	int t = lua_gettop(L);
	if (!type) {
		lua_pushstring(L, value);
		return;
	}
	/* Check the Oid cache for 'type'. */
	lua_pushnumber(L, (double)type);
	lua_pushstring(L, "lpg_oid_cache");
	lua_rawget(L, LUA_ENVIRONINDEX);
	lua_pushvalue(L, t + 1);
	lua_rawget(L, -2);
	if (lua_type(L, -1) != LUA_TSTRING) {
		/* Cache miss. Look up the type string in the DB. */
		PGresult *res;
		char *stmt = "SELECT typname FROM pg_type WHERE oid = $1::oid";
		const char *oid;
		lua_pop(L, 1);
		lua_pushvalue(L, t + 1);
		oid = lua_tostring(L, -1);
		res = PQexecParams(pg_conn, stmt, 1, NULL, &oid, NULL, NULL, 0);
		lua_pop(L, 1);
		if (res && PQresultStatus(res) == PGRES_TUPLES_OK &&
			PQntuples(res) == 1 && !PQgetisnull(res, 0, 0)) {
			/* A type string was found. */
			lua_pushstring(L, PQgetvalue(res, 0, 0));
			lua_pushvalue(L, t + 1);
			/* Cache the type string and Oid. */
			lua_pushvalue(L, -2);
			lua_pushvalue(L, -2);
			lua_rawset(L, t + 2);
			lua_pushvalue(L, -2);
			lua_rawset(L, t + 2);
			PQclear(res);
		} else {
			/* A type string was not found. */
			lua_settop(L, t);
			lua_pushstring(L, value);
			PQclear(res);
			return;
		}
	}
	lua_remove(L, -2);
	lua_remove(L, -2);
	/* Check for a custom type parser. */
	lua_pushstring(L, "lpg_type_parser_over");
	lua_rawget(L, LUA_ENVIRONINDEX);
	lua_pushvalue(L, -2);
	lua_rawget(L, -2);
	if (luaL_getmetafield(L, -1, "__call")) {
		/*
		 * Custom type parser found (object with a __call metamethod).
		 */
		lua_replace(L, t + 1);
		lua_replace(L, t + 2);
		lua_pushstring(L, value);
		lua_call(L, 2, 1);
	} else {
		if (!lua_isfunction(L, -1)) {
			lua_pop(L, 2);
			/* Check for a built in type parser. */
			lua_pushstring(L, "lpg_type_parser");
			lua_rawget(L, LUA_ENVIRONINDEX);
			lua_pushvalue(L, -2);
			lua_rawget(L, -2);
			if (!lua_isfunction(L, -1)) {
				/* A type parser was not found. */
				lua_settop(L, t);
				lua_pushstring(L, value);
				return;
			}
		}
		/* Custom or built in type parser found (function). */
		lua_remove(L, -2);
		lua_remove(L, -2);
		lua_pushstring(L, value);
		lua_call(L, 1, 1);
	}
}


/*
 * Notice processor for the PGconn.
 */
static void lpg_notice_processor(void *arg, const char *message)
{
	return;
}


/*
 * Validate a custom handler.
 * The top of the stack is either a function or an object with a __call
 * metamethod. Directly bellow the top of the stack is the name of the custom
 * handler. Check that the handler does not already exist in the table at idx.
 */
static void lpg_check_handler(lua_State *L, int idx)
{
	if (lua_type(L, -2) != LUA_TSTRING)
		luaL_argerror(L, 1, "String expected.");
	if (luaL_getmetafield(L, -1, "__call") && lua_isfunction(L, -1))
		lua_pop(L, 1);
	else if (!lua_isfunction(L, -1))
		luaL_argerror(L, 2, "Function expected.");
	lua_pushvalue(L, -2);
	lua_rawget(L, idx);
	if (!lua_isnil(L, -1))
		luaL_argerror(L, 1, "Handler name already exists.");
	lua_pop(L, 1);
}


/*
 * Set a custom type parameter.
 * Arguments:
 *	1. Param number. Must be greater than zero and less than or equal to
 *		the current number of parameters plus one.
 *	2 .. n. Optional arguments sent to the custom type formatter.
 */
static int lpg_set_type(lua_State *L)
{
	struct lpg_stmt *stmt;
	PGconn *pg_conn;
	int i, j;
	stmt = lpg_checkstmt(L, 1, 1);
	i = luaL_checkint(L, 2);
	j = stmt->params + 1;
	luaL_argcheck(L, (0 < i && i <= j), 2, "Index out of range.");
	lua_rawgeti(L, LUA_ENVIRONINDEX, stmt->conn);
	pg_conn = *lpg_checkconn(L, lua_gettop(L), 1);
	lua_pop(L, 1);
	/* Get the type formatter. */
	lua_pushvalue(L, lua_upvalueindex(1));
	if (luaL_getmetafield(L, -1, "__call")) {
		lua_insert(L, 3);
		lua_insert(L, 4);
	} else
		lua_insert(L, 3);
	/*
	 * Call the type formatter. Result should be a string and a type name.
	 */
	lua_call(L, lua_gettop(L) - 3, 2);
	/* Convert the type name to an Oid or nil. */
	lpg_getoid(L, pg_conn);
	/* Add the paramter and its type to the param table. */
	lua_getfenv(L, 1);
	lua_pushvalue(L, 3);
	lua_rawseti(L, 5, i);
	lua_pushvalue(L, 4);
	lua_rawseti(L, 5, -i);
	if (i == j)
		stmt->params = j;
	return 0;
}


/****************************************************************************/


/*
 * Connect to a database.
 * The connection info string is the only argument.
 * Returns a PGconn userdata on success, or nil plus an error message.
 */
static int lpg_lib_connect(lua_State *L)
{
	PGconn *pg_conn = PQconnectdb(luaL_checkstring(L, 1));
	if (PQstatus(pg_conn) == CONNECTION_OK) {
		PQsetNoticeProcessor(pg_conn, lpg_notice_processor, NULL);
		lpg_newconn(L, pg_conn);
		return 1;
	} else {
		PQfinish(pg_conn);
		lua_pushnil(L);
		lua_pushstring(L, "Error connecting to database.");
		return 2;
	}
}


/*
 * Install a custom type formatter.
 * Accepts either a single type formatter, or a table of type formatters.
 */
static int lpg_lib_type_format(lua_State *L)
{
	lua_settop(L, 2);
	lua_pushstring(L, "lpg_stmt_metatable");
	lua_rawget(L, LUA_ENVIRONINDEX);
	lua_pushstring(L, "__index");
	lua_rawget(L, -2);
	lua_remove(L, -2);
	if (lua_istable(L, 1)) {
		lua_remove(L, -2);
		lua_pushnil(L);
		while (lua_next(L, 1)) {
			lpg_check_handler(L, 2);
			lua_pushvalue(L, -2);
			lua_pushvalue(L, -2);
			lua_pushcclosure(L, lpg_set_type, 1);
			lua_rawset(L, 2);
			lua_pop(L, 1);
		}
	} else {
		lua_insert(L, 1);
		lpg_check_handler(L, 1);
		lua_pushcclosure(L, lpg_set_type, 1);
		lua_rawset(L, 1);
	}
	return 0;
}


/*
 * Install a custom type parser.
 * Accepts either a single type parser, or a table of type parsers.
 */
static int lpg_lib_type_parser(lua_State *L)
{
	lua_settop(L, 2);
	lua_pushstring(L, "lpg_type_parser_over");
	lua_rawget(L, LUA_ENVIRONINDEX);
	if (lua_istable(L, 1)) {
		lua_remove(L, -2);
		lua_pushnil(L);
		while (lua_next(L, 1)) {
			lpg_check_handler(L, 2);
			lua_pushvalue(L, -2);
			lua_pushvalue(L, -2);
			lua_rawset(L, 2);
			lua_pop(L, 1);
		}
	} else {
		lua_insert(L, 1);
		lpg_check_handler(L, 1);
		lua_rawset(L, 1);
	}
	return 0;
}


/****************************************************************************/


/*
 * Close the PGconn.
 */
static int lpg_conn_close(lua_State *L)
{
	PGconn **conn;
	lua_settop(L, 1);
	conn = lpg_checkconn(L, 1, 0);
	if (*conn) {
		PQfinish(*conn);
		*conn = NULL;
	}
	return 0;
}


/*
 * Execute an SQL statment and discard any results.
 * Returns the number of rows affected, the number of rows returned, or nil
 * plus an error message.
 */
static int lpg_conn_execute(lua_State *L)
{
	PGconn **conn;
	PGresult *res;
	const char *statement;
	lua_settop(L, 2);
	conn = lpg_checkconn(L, 1, 1);
	statement = luaL_checkstring(L, 2);
	res = lpg_execute(L, *conn, statement, 0, 0);
	return lpg_res_clear(L, res);
}


/*
 * Begin a transaction.
 */
static int lpg_conn_begin(lua_State *L)
{
	lua_settop(L, 1);
	lua_pushstring(L, "BEGIN");
	return lpg_conn_execute(L);
}


/*
 * Commit a transaction.
 */
static int lpg_conn_commit(lua_State *L)
{
	lua_settop(L, 1);
	lua_pushstring(L, "COMMIT");
	return lpg_conn_execute(L);
}


/*
 * Rollback a transaction.
 */
static int lpg_conn_rollback(lua_State *L)
{
	lua_settop(L, 1);
	lua_pushstring(L, "ROLLBACK");
	return lpg_conn_execute(L);
}


/*
 * Create a new statement userdata.
 */
static int lpg_conn_statement(lua_State *L)
{
	lua_settop(L, 2);
	lpg_checkconn(L, 1, 1);
	luaL_checkstring(L, 2);
	lpg_newstmt(L, 1, 2);
	return 1;
}


/*
 * Return a string describing the PGconn.
 */
static int lpg_conn_tostring(lua_State *L)
{
	PGconn **conn;
	lua_settop(L, 1);
	conn = lpg_checkconn(L, 1, 0);
	if (*conn)
		lua_pushfstring(L, "PostgreSQL connection (%p)", *conn);
	else
		lua_pushstring(L, "PostgreSQL connection (closed)");
	return 1;
}


/****************************************************************************/


/*
 * Close the statement.
 */
static int lpg_stmt_close(lua_State *L)
{
	struct lpg_stmt *stmt;
	lua_settop(L, 1);
	stmt = lpg_checkstmt(L, 1, 0);
	if (!stmt->closed) {
		luaL_unref(L, LUA_ENVIRONINDEX, stmt->conn);
		stmt->conn = LUA_NOREF;
		stmt->closed = 1;
	}
	return 0;
}


/*
 * Reduce the number of parameters for the statement.
 * The new number must be greater than zero and less than or equal to the
 * current number.
 */
static int lpg_stmt_set_params(lua_State *L)
{
	struct lpg_stmt *stmt;
	int i, j;
	lua_settop(L, 2);
	stmt = lpg_checkstmt(L, 1, 1);
	i = luaL_checkint(L, 2);
	j = stmt->params;
	luaL_argcheck(L, (0 <= i && i <= j) , 2, "Index out of range.");
	if (i < j) {
		lua_getfenv(L, 1);
		while (i < j) {
			lua_pushnil(L);
			lua_rawseti(L, -2, j);
			lua_pushnil(L);
			lua_rawseti(L, -2, -j);
			j--;
		}
		lua_pop(L, 1);
	}
	stmt->params = i;
	return 0;
}


/*
 * Return the number of parameters for the statement.
 */
static int lpg_stmt_params(lua_State *L)
{
	struct lpg_stmt *stmt;
	lua_settop(L, 1);
	stmt = lpg_checkstmt(L, 1, 1);
	lua_pushinteger(L, stmt->params);
	return 1;
}


/*
 * Set a string parameter with a given type.
 * Arguments:
 *	1. Param number. Must be greater than zero and less than or equal to
 *		the current number of parameters plus one.
 *	2. The parameter. Must be a string.
 *	3. The param type. Must be a string, defaults to 'text'.
 */
static int lpg_stmt_set_string(lua_State *L)
{
	struct lpg_stmt *stmt;
	PGconn *pg_conn;
	int i, j;
	lua_settop(L, 4);
	stmt = lpg_checkstmt(L, 1, 1);
	i = luaL_checkint(L, 2);
	j = stmt->params + 1;
	luaL_argcheck(L, (0 < i && i <= j), 2, "Index out of range.");
	luaL_checkstring(L, 3);
	if (lua_type(L, 4) != LUA_TSTRING) {
		lua_pop(L, 1);
		lua_pushstring(L, "text");
	}
	lua_rawgeti(L, LUA_ENVIRONINDEX, stmt->conn);
	pg_conn = *lpg_checkconn(L, lua_gettop(L), 1);
	lua_pop(L, 1);
	/* Convert the type name to an Oid or nil. */
	lpg_getoid(L, pg_conn);
	/* Add the paramter and its type to the param table. */
	lua_getfenv(L, 1);
	lua_pushvalue(L, 3);
	lua_rawseti(L, 5, i);
	lua_pushvalue(L, 4);
	lua_rawseti(L, 5, -i);
	if (i == j)
		stmt->params = j;
	return 0;
}


/*
 * Set a floating point type number parameter.
 * Arguments:
 *	1. Param number. Must be greater than zero and less than or equal to
 *		the current number of parameters plus one.
 *	2. The parameter. Must be a number.
 *	3. Precision. Either 4 or 8 bytes, default is 8 bytes.
 */
static int lpg_stmt_set_number(lua_State *L)
{
	struct lpg_stmt *stmt;
	PGconn *pg_conn;
	int i, j, s;
	char buff[32];
	double n;
	lua_settop(L, 4);
	stmt = lpg_checkstmt(L, 1, 1);
	i = luaL_checkint(L, 2);
	j = stmt->params + 1;
	luaL_argcheck(L, (0 < i && i <= j), 2, "Index out of range.");
	n = luaL_checknumber(L, 3);
	s = luaL_optint(L, 4, 8);
	switch (s) {
	default:
		s = 8;
	case 8:
		sprintf(buff, "%.16g", n);
		break;
	case 4:
		sprintf(buff, "%.8g", (float)n);
		break;
	}
	lua_pushstring(L, buff);
	lua_pushfstring(L, "float%d", s);
	lua_rawgeti(L, LUA_ENVIRONINDEX, stmt->conn);
	pg_conn = *lpg_checkconn(L, lua_gettop(L), 1);
	lua_pop(L, 1);
	/* Convert the type name to an Oid or nil. */
	lpg_getoid(L, pg_conn);
	/* Add the paramter and its type to the param table. */
	lua_getfenv(L, 1);
	lua_pushvalue(L, 5);
	lua_rawseti(L, 7, i);
	lua_pushvalue(L, 6);
	lua_rawseti(L, 7, -i);
	if (i == j)
		stmt->params = j;
	return 0;
}


/*
 * Set a integer type parameter.
 * Arguments:
 *	1. Param number. Must be greater than zero and less than or equal to
 *		the current number of parameters plus one.
 *	2. The parameter. Must be an integer.
 *	3. Precision. Either 2, 4 or 8 bytes, default is 4 bytes.
 */
static int lpg_stmt_set_integer(lua_State *L)
{
	struct lpg_stmt *stmt;
	PGconn *pg_conn;
	int i, j, s;
	char buff[32];
	lua_settop(L, 4);
	stmt = lpg_checkstmt(L, 1, 1);
	i = luaL_checkint(L, 2);
	j = stmt->params + 1;
	luaL_argcheck(L, (0 < i && i <= j), 2, "Index out of range.");
	s = luaL_optint(L, 4, 4);
	switch (s) {
	case 8:
		sprintf(buff, "%lld", (long long int)luaL_checknumber(L, 3));
		break;
	default:
		s = 4;
	case 4:
		sprintf(buff, "%d", luaL_checkint(L, 3));
		break;
	case 2:
		sprintf(buff, "%hd", (short int)luaL_checkint(L, 3));
		break;
	}
	lua_pushstring(L, buff);
	lua_pushfstring(L, "int%d", s);
	lua_rawgeti(L, LUA_ENVIRONINDEX, stmt->conn);
	pg_conn = *lpg_checkconn(L, lua_gettop(L), 1);
	lua_pop(L, 1);
	/* Convert the type name to an Oid or nil. */
	lpg_getoid(L, pg_conn);
	/* Add the paramter and its type to the param table. */
	lua_getfenv(L, 1);
	lua_pushvalue(L, 5);
	lua_rawseti(L, 7, i);
	lua_pushvalue(L, 6);
	lua_rawseti(L, 7, -i);
	if (i == j)
		stmt->params = j;
	return 0;
}


/*
 * Set a boolean type parameter.
 * Arguments:
 *	1. Param number. Must be greater than zero and less than or equal to
 *		the current number of parameters plus one.
 *	2. The parameter. Any type.
 */
static int lpg_stmt_set_boolean(lua_State *L)
{
	struct lpg_stmt *stmt;
	PGconn *pg_conn;
	int i, j;
	lua_settop(L, 3);
	stmt = lpg_checkstmt(L, 1, 1);
	i = luaL_checkint(L, 2);
	j = stmt->params + 1;
	luaL_argcheck(L, (0 < i && i <= j), 2, "Index out of range.");
	lua_pushstring(L, lua_toboolean(L, 3) ? "t" : "f");
	lua_pushstring(L, "boolean");
	lua_rawgeti(L, LUA_ENVIRONINDEX, stmt->conn);
	pg_conn = *lpg_checkconn(L, lua_gettop(L), 1);
	lua_pop(L, 1);
	/* Convert the type name to an Oid or nil. */
	lpg_getoid(L, pg_conn);
	/* Add the paramter and its type to the param table. */
	lua_getfenv(L, 1);
	lua_pushvalue(L, 4);
	lua_rawseti(L, 6, i);
	lua_pushvalue(L, 5);
	lua_rawseti(L, 6, -i);
	if (i == j)
		stmt->params = j;
	return 0;
}


/*
 * Set a null parameter. This does not reduce the number of parameters sent.
 * Arguments:
 *	1. Param number. Must be greater than zero and less than or equal to
 *		the current number of parameters plus one.
 */
static int lpg_stmt_set_null(lua_State *L)
{
	struct lpg_stmt *stmt;
	int i, j;
	lua_settop(L, 2);
	stmt = lpg_checkstmt(L, 1, 1);
	i = luaL_checkint(L, 2);
	j = stmt->params + 1;
	luaL_argcheck(L, (0 < i && i <= j), 2, "Index out of range.");
	/* Add the paramter and its type to the param table. */
	lua_getfenv(L, 1);
	lua_pushnil(L);
	lua_rawseti(L, 3, i);
	lua_pushnil(L);
	lua_rawseti(L, 3, -i);
	if (i == j)
		stmt->params = j;
	return 0;
}


/*
 * Execute the statement and discard any results.
 * Returns the number of rows affected, the number of rows returned, or nil
 * plus an error message.
 */
static int lpg_stmt_execute(lua_State *L)
{
	struct lpg_stmt *stmt;
	PGconn *pg_conn;
	const char *statement;
	PGresult *res;
	lua_settop(L, 1);
	stmt = lpg_checkstmt(L, 1, 1);
	lua_rawgeti(L, LUA_ENVIRONINDEX, stmt->conn);
	pg_conn = *lpg_checkconn(L, 2, 1);
	lua_getfenv(L, 1);
	lua_pushstring(L, "statement");
	lua_rawget(L, 3);
	statement = lua_tostring(L, 4);
	res = lpg_execute(L, pg_conn, statement, stmt->params, 3);
	return lpg_res_clear(L, res);
}


/*
 * Execute the statement and return only the first row of results.
 * Returns a table containing the first row of data plus the total number
 * of rows returned by the statement. If there was an error, nil plus an
 * error message are returned.
 * Arguments:
 *	1. An optional table to put the result data into. This will overwrite
 *		any data already in the table. Note: weird things can happen
 *		if the result data contains nulls and this table is non-empty.
 */
static int lpg_stmt_select_result(lua_State *L)
{
	struct lpg_stmt *stmt;
	PGconn *pg_conn;
	const char *statement;
	PGresult *res;
	lua_settop(L, 2);
	stmt = lpg_checkstmt(L, 1, 1);
	if (!lua_istable(L, 2)) {
		lua_pop(L, 1);
		lua_createtable(L, 0, 0);
	}
	lua_rawgeti(L, LUA_ENVIRONINDEX, stmt->conn);
	pg_conn = *lpg_checkconn(L, 3, 1);
	lua_getfenv(L, 1);
	lua_pushstring(L, "statement");
	lua_rawget(L, 4);
	statement = lua_tostring(L, 5);
	res = lpg_execute(L, pg_conn, statement, stmt->params, 4);
	lua_pop(L, 2);
	if (res && PQresultStatus(res) == PGRES_TUPLES_OK) {
		int tuples = PQntuples(res);
		int fields = PQnfields(res);
		int r = 0, f;
		/* If there is at least one row of results: */
		if (tuples > 0) {
			/* Push each field into the row table. */
			for (f = 0; f < fields; f++) {
				lua_pushstring(L, PQfname(res, f));
				if (PQgetisnull(res, r, f))
					lua_pushnil(L);
				else
					lpg_pushtype(L, pg_conn,
							PQftype(res, f),
							PQgetvalue(res, r, f));
				lua_rawset(L, 2);
			}
		}
		lua_pushvalue(L, 2);
		lua_pushinteger(L, tuples);
		PQclear(res);
		return 2;
	} else
		return lpg_res_clear(L, res);
}


/*
 * Execute the statement and return the entire result set.
 * Returns a table containing each row of data, plus the total number of
 * rows returned by the statement. If there was an error, nil plus an
 * error message are returned.
 * Arguments:
 *	1. An optional table to put the result data into. If the table already
 *		contains data, the new results from the query are appended to
 *		the end of the table.
 */
static int lpg_stmt_select_result_set(lua_State *L)
{
	struct lpg_stmt *stmt;
	PGconn *pg_conn;
	const char *statement;
	PGresult *res;
	lua_settop(L, 2);
	stmt = lpg_checkstmt(L, 1, 1);
	if (!lua_istable(L, 2)) {
		lua_pop(L, 1);
		lua_createtable(L, 0, 0);
	}
	lua_rawgeti(L, LUA_ENVIRONINDEX, stmt->conn);
	pg_conn = *lpg_checkconn(L, 3, 1);
	lua_getfenv(L, 1);
	lua_pushstring(L, "statement");
	lua_rawget(L, 4);
	statement = lua_tostring(L, 5);
	res = lpg_execute(L, pg_conn, statement, stmt->params, 4);
	lua_pop(L, 2);
	if (res && PQresultStatus(res) == PGRES_TUPLES_OK) {
		int tuples = PQntuples(res);
		int fields = PQnfields(res);
		int i = lua_objlen(L, 2);
		int r, f;
		/* For each row in the result: */
		for (r = 0; r < tuples; r++) {
			lua_createtable(L, 0, fields);
			/* Push each field into the row table. */
			for (f = 0; f < fields; f++) {
				lua_pushstring(L, PQfname(res, f));
				if (PQgetisnull(res, r, f))
					lua_pushnil(L);
				else
					lpg_pushtype(L, pg_conn,
							PQftype(res, f),
							PQgetvalue(res, r, f));
				lua_rawset(L, 4);
			}
			/* Push the row table onto the result table. */
			lua_rawseti(L, 2, ++i);
		}
		lua_pushvalue(L, 2);
		lua_pushinteger(L, tuples);
		PQclear(res);
		return 2;
	} else
		return lpg_res_clear(L, res);
}


/*
 * Execute the statement and return only the data from the first row and the
 * specified column.
 * Returns a single result, plus the total number of rows returned by the
 * statement. If there was an error, nil plus an error message are returned.
 * Arguments:
 *	1. Field number. A valid column number for the result (1 .. n).
 *		Default is 1.
 */
static int lpg_stmt_select_field(lua_State *L)
{
	struct lpg_stmt *stmt;
	PGconn *pg_conn;
	const char *statement;
	int field;
	PGresult *res;
	lua_settop(L, 2);
	stmt = lpg_checkstmt(L, 1, 1);
	field = luaL_optint(L, 2, 1) - 1;
	lua_rawgeti(L, LUA_ENVIRONINDEX, stmt->conn);
	pg_conn = *lpg_checkconn(L, 3, 1);
	lua_getfenv(L, 1);
	lua_pushstring(L, "statement");
	lua_rawget(L, 4);
	statement = lua_tostring(L, 5);
	res = lpg_execute(L, pg_conn, statement, stmt->params, 4);
	lua_pop(L, 2);
	if (res && PQresultStatus(res) == PGRES_TUPLES_OK) {
		int tuples = PQntuples(res);
		int fields = PQnfields(res);
		/* If there is at least one row of results: */
		if (tuples > 0 && 0 <= field && field < fields &&
			!PQgetisnull(res, 0, field))
			/* Select the desired field from the first row. */
			lpg_pushtype(L, pg_conn, PQftype(res, field),
					PQgetvalue(res, 0, field));
		else
			lua_pushnil(L);
		lua_pushinteger(L, tuples);
		PQclear(res);
		return 2;
	} else
		return lpg_res_clear(L, res);
}


/*
 * Execute the statement and return the first two columns of data as
 * key-value paired table.
 * Returns a table containing each row of data, plus the total number of
 * rows returned by the statement. If there was an error, nil plus an
 * error message are returned.
 */
static int lpg_stmt_select_pairs(lua_State *L)
{
	struct lpg_stmt *stmt;
	PGconn *pg_conn;
	const char *statement;
	PGresult *res;
	lua_settop(L, 1);
	stmt = lpg_checkstmt(L, 1, 1);
	lua_rawgeti(L, LUA_ENVIRONINDEX, stmt->conn);
	pg_conn = *lpg_checkconn(L, 2, 1);
	lua_getfenv(L, 1);
	lua_pushstring(L, "statement");
	lua_rawget(L, 3);
	statement = lua_tostring(L, 4);
	res = lpg_execute(L, pg_conn, statement, stmt->params, 3);
	lua_pop(L, 2);
	if (res && PQresultStatus(res) == PGRES_TUPLES_OK) {
		int tuples = PQntuples(res);
		int r;
		lua_createtable(L, 0, 0);
		/* For each row in the result: */
		for (r = 0; r < tuples; r++) {
			/*
			 * Push the key-valued pair (col0-col1) on to the
			 * result table. Skip if the key is nil.
			 */
			if (!PQgetisnull(res, r, 0)) {
				lpg_pushtype(L, pg_conn, PQftype(res, 0),
						PQgetvalue(res, r, 0));
				if (PQgetisnull(res, r, 1))
					lua_pushnil(L);
				else
					lpg_pushtype(L, pg_conn,
							PQftype(res, 1),
							PQgetvalue(res, r, 1));
				lua_rawset(L, 3);
			}
		}
		lua_pushinteger(L, tuples);
		PQclear(res);
		return 2;
	} else
		return lpg_res_clear(L, res);
}


/*
 * Execute the statement and call a handler function for each row of data
 * in the result.
 * Returns the total number of rows returned by the statement. If there was
 * an error, nil plus an error message are returned.
 * Arguments:
 *	1. A handler function. This function will be called with one row
 *		of data at time, wraped in a table. If the function returns
 *		false or nil execution will stop and any remaining rows will
 *		be discarded.
 */
static int lpg_stmt_select_handler(lua_State *L)
{
	struct lpg_stmt *stmt;
	PGconn *pg_conn;
	const char *statement;
	PGresult *res;
	lua_settop(L, 2);
	stmt = lpg_checkstmt(L, 1, 1);
	luaL_checktype(L, 2, LUA_TFUNCTION);
	lua_rawgeti(L, LUA_ENVIRONINDEX, stmt->conn);
	pg_conn = *lpg_checkconn(L, 3, 1);
	lua_getfenv(L, 1);
	lua_pushstring(L, "statement");
	lua_rawget(L, 4);
	statement = lua_tostring(L, 5);
	res = lpg_execute(L, pg_conn, statement, stmt->params, 4);
	lua_pop(L, 2);
	if (res && PQresultStatus(res) == PGRES_TUPLES_OK) {
		int tuples = PQntuples(res);
		int fields = PQnfields(res);
		int r, f;
		/* For each row in the result: */
		for (r = 0; r < tuples; r++) {
			lua_pushvalue(L, 2);
			lua_createtable(L, 0, fields);
			/* Push each field into the row table. */
			for (f = 0; f < fields; f++) {
				lua_pushstring(L, PQfname(res, f));
				if (PQgetisnull(res, r, f))
					lua_pushnil(L);
				else
					lpg_pushtype(L, pg_conn,
							PQftype(res, f),
							PQgetvalue(res, r, f));
				lua_rawset(L, 5);
			}
			/* Call the handler function with the row table. */
			lua_call(L, 1, 1);
			/* If it returns false or nil, we are done. */
			if (lua_isnil(L, 4) || 
				(lua_isboolean(L, 4) && !lua_toboolean(L, 4)))
				break;
			lua_pop(L, 1);
		}
		lua_pushinteger(L, tuples);
		PQclear(res);
		return 1;
	} else
		return lpg_res_clear(L, res);
}


/*
 * Return a string describing the statement.
 * If the statement is still open, it will be printed out as well,
 * along with all of its parameters.
 */
static int lpg_stmt_tostring(lua_State *L)
{
	struct lpg_stmt *stmt;
	lua_settop(L, 1);
	stmt = lpg_checkstmt(L, 1, 0);
	if (stmt->closed)
		lua_pushstring(L, "PostgreSQL statement (closed)");
	else {
		int i;
		lua_getfenv(L, 1);
		lua_pushfstring(L, "PostgreSQL statement (%p)\n", stmt);
		lua_pushstring(L, "\n");
		lua_getfield(L, 2, "statement");
		lua_pushvalue(L, 4);
		for (i = 1; i <= stmt->params; i++) {
			lua_pushfstring(L, "$%d = ", i);
			lua_rawgeti(L, 2, i);
			lua_pushvalue(L, 4);
		}
		lua_concat(L, 4 + 3 * stmt->params);
		return 1;
	}
	return 1;
}


/****************************************************************************/


/*
 * Parse a number argument.
 */
static int lpg_type_parser_number(lua_State *L)
{
	lua_settop(L, 1);
	if (lua_isnumber(L, 1))
		lua_pushnumber(L, lua_tonumber(L, 1));
	else
		lua_pushnil(L);
	return 1;
}


/*
 * Parse a string argument.
 */
static int lpg_type_parser_string(lua_State *L)
{
	lua_settop(L, 1);
	if (lua_type(L, 1) != LUA_TSTRING)
		lua_pushnil(L);
	return 1;
}


/*
 * Parse a boolean argument.
 */
static int lpg_type_parser_boolean(lua_State *L)
{
	lua_settop(L, 1);
	if (lua_type(L, 1) == LUA_TSTRING) {
		switch (*lua_tostring(L, 1)) {
		case 'T':
		case 't':
		case 'y':
		case '1':
			lua_pushboolean(L, 1);
			break;
		case 'F':
		case 'f':
		case 'n':
		case '0':
			lua_pushboolean(L, 0);
			break;
		default:
			lua_pushnil(L);
			break;
		}
	} else
		lua_pushnil(L);
	return 1;
}


/****************************************************************************/


/*
 * Library methods.
 */
static const struct luaL_Reg lpg_lib[] = {
	{"connect",		lpg_lib_connect},
	{"typeFormat",		lpg_lib_type_format},
	{"typeParser",		lpg_lib_type_parser},
	{NULL, NULL}
};


/*
 * PGconn method table.
 */
static const struct luaL_Reg lpg_conn_method_table[] = {
	{"close",		lpg_conn_close},
	{"begin",		lpg_conn_begin},
	{"commit",		lpg_conn_commit},
	{"rollback",		lpg_conn_rollback},
	{"execute",		lpg_conn_execute},
	{"statement",		lpg_conn_statement},
	{NULL, NULL}
};


/*
 * PGconn metatable.
 */
static const struct luaL_Reg lpg_conn_metatable[] = {
	{"__gc",		lpg_conn_close},
	{"__call",		lpg_conn_execute},
	{"__tostring",		lpg_conn_tostring},
	{NULL, NULL}
};


/*
 * Statement method table.
 */
static const struct luaL_Reg lpg_stmt_method_table[] = {
	{"close",		lpg_stmt_close},
	{"setParams",		lpg_stmt_set_params},
	{"setString",		lpg_stmt_set_string},
	{"setNumber",		lpg_stmt_set_number},
	{"setInteger",		lpg_stmt_set_integer},
	{"setBoolean",		lpg_stmt_set_boolean},
	{"setNull",		lpg_stmt_set_null},
	{"execute",		lpg_stmt_execute},
	{"selectResult",	lpg_stmt_select_result},
	{"selectResultSet",	lpg_stmt_select_result_set},
	{"selectField",		lpg_stmt_select_field},
	{"selectPairs",		lpg_stmt_select_pairs},
	{"selectHandler",	lpg_stmt_select_handler},
	{NULL, NULL}
};


/*
 * Statement metatable.
 */
static const struct luaL_Reg lpg_stmt_metatable[] = {
	{"__gc",		lpg_stmt_close},
	{"__len",		lpg_stmt_params},
	{"__call",		lpg_stmt_execute},
	{"__tostring",		lpg_stmt_tostring},
	{NULL, NULL}
};


/*
 * Built in type parsers.
 */
static const struct luaL_Reg lpg_type_parser[] = {
	{"int2",		lpg_type_parser_number},
	{"int4",		lpg_type_parser_number},
	{"int8",		lpg_type_parser_number},
	{"float4",		lpg_type_parser_number},
	{"float8",		lpg_type_parser_number},
	{"numeric",		lpg_type_parser_number},
	{"varchar",		lpg_type_parser_string},
	{"char",		lpg_type_parser_string},
	{"text",		lpg_type_parser_string},
	{"boolean",		lpg_type_parser_boolean},
	{NULL, NULL}
};


/*
 * Create a new metatable and an associated method table.
 */
static void lpg_create_metatable(lua_State *L, const luaL_reg *metatable,
					const luaL_reg *method_table)
{
	lua_createtable(L, 0, 2);

	lua_pushstring(L, "__index");
	lua_createtable(L, 0, 0);
	luaL_register(L, NULL, method_table);
	lua_rawset(L, -3);

	lua_pushstring(L, "__metatable");
	lua_pushstring(L, "PostgreSQL private metatable.");
	lua_rawset(L, -3);

	luaL_register(L, NULL, metatable);
}


/*
 * Setup the new environment for the library. Register the driver.
 */
LUALIB_API int luaopen_postgres(lua_State *L)
{
	lua_settop(L, 1);

	lua_createtable(L, 0, 5);

	lua_pushvalue(L, 2);
	lua_replace(L, LUA_ENVIRONINDEX);

	lua_pushstring(L, "lpg_conn_metatable");
	lpg_create_metatable(L, lpg_conn_metatable, lpg_conn_method_table);
	lua_rawset(L, 2);

	lua_pushstring(L, "lpg_stmt_metatable");
	lpg_create_metatable(L, lpg_stmt_metatable, lpg_stmt_method_table);
	lua_rawset(L, 2);

	lua_pushstring(L, "lpg_type_parser");
	lua_createtable(L, 0, 0);
	luaL_register(L, NULL, lpg_type_parser);
	lua_rawset(L, 2);

	lua_pushstring(L, "lpg_type_parser_over");
	lua_createtable(L, 0, 0);
	lua_rawset(L, 2);

	lua_pushstring(L, "lpg_oid_cache");
	lua_createtable(L, 0, 0);
	lua_rawset(L, 2);

	luaL_register(L, luaL_optstring(L, 1, "postgres"), lpg_lib);

	return 1;
}

